//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd-tor.rs: Syd's SOCKS Proxy Forwarder
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon socksns crate which is:
//     Copyright (c) 2020 Steven Engler
//     SPDX-License-Identifier: MIT
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    collections::HashMap,
    env,
    ffi::CString,
    net::{IpAddr, Ipv4Addr, SocketAddr},
    os::{
        fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
        unix::net::UnixStream,
    },
    path::PathBuf,
    process::{exit, ExitCode},
};

use anyhow::{bail, Context, Result};
use getargs::{Opt, Options};
use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
use nix::{
    errno::Errno,
    fcntl::{splice, OFlag, SpliceFFlags},
    libc::PIPE_BUF,
    sys::{
        epoll::{
            epoll_create1, epoll_ctl, epoll_wait, EpollCreateFlags, EpollEvent, EpollFlags, EpollOp,
        },
        resource::{getrlimit, setrlimit, Resource},
        signal::Signal,
        socket::{
            accept4, connect, socket, AddressFamily, SockFlag, SockType, SockaddrIn, SockaddrIn6,
            SockaddrLike,
        },
    },
    unistd::{close, pipe2, write},
};
use sendfd::RecvWithFd;
use syd::fs::{path_exists, set_nonblock};

#[allow(clippy::disallowed_methods)]
fn main() -> Result<ExitCode> {
    // Parse options.
    let (fpid, sock, addr, debug) = parse_options()?;

    // Ignore all signals except SIG{KILL,STOP,SYS}.
    // Ignore errors, not much we can do about them.
    for signal in Signal::iterator() {
        if !matches!(signal, Signal::SIGKILL | Signal::SIGSTOP | Signal::SIGSYS) {
            let _ = syd::ignore_signal(signal);
        }
    }

    // Attempt to set file-max to hard limit overriding the soft limit.
    if let Ok((soft_limit, hard_limit)) = getrlimit(Resource::RLIMIT_NOFILE) {
        if soft_limit < hard_limit {
            let _ = setrlimit(Resource::RLIMIT_NOFILE, hard_limit, hard_limit);
        }
    }

    // Receive the socket bound inside the namespace.
    let mut buf = [0u8; 1];
    let mut lfd = [-1; 1];
    let proxy = match sock.recv_with_fd(&mut buf, &mut lfd) {
        Ok((_, 1)) if lfd[0] >= 0 => {
            // Socket no longer necessary.
            drop(sock);
            // SAFETY: recvmsg received a valid FD.
            unsafe { OwnedFd::from_raw_fd(lfd[0]) }
        }
        Ok(_) => bail!("Didn't receive the fd on the socket"),
        Err(error) => {
            return Err(anyhow::Error::msg(error).context("Didn't receive the fd on the socket"))
        }
    };

    // Create epoll instance.
    let epoll_fd = epoll_create1(EpollCreateFlags::EPOLL_CLOEXEC)
        .map_err(anyhow::Error::msg)
        .context("Failed to create epoll instance")?;

    // Confine.
    confine(&*addr, debug).context("Failed to confine")?;

    let debug_fd = if debug { nix::libc::STDERR_FILENO } else { -42 };

    // Run the proxy server.
    match run_proxy_server(epoll_fd, fpid.as_raw_fd(), proxy.as_raw_fd(), addr) {
        Ok(_) => Ok(ExitCode::SUCCESS),
        Err(error) => {
            let error = format!("syd-tor: {error}\n");
            let error = CString::new(error).unwrap();
            let _ = write(debug_fd, error.as_bytes_with_nul());
            Ok(ExitCode::FAILURE)
        }
    }
}

/// Run the proxy server
pub fn run_proxy_server(
    epoll_fd: RawFd,
    pid_fd: RawFd,
    listening_fd: RawFd,
    external_addr: Box<dyn SockaddrLike>,
) -> Result<()> {
    // Set listening fd non blocking.
    set_nonblock(&listening_fd, true)
        .map_err(anyhow::Error::msg)
        .context("Failed to set listen fd nonblocking")?;

    // Add PIDFd to epoll (becomes readable when process terminates).
    let mut event = EpollEvent::new(
        EpollFlags::EPOLLET | EpollFlags::EPOLLIN | EpollFlags::EPOLLHUP | EpollFlags::EPOLLRDHUP,
        pid_fd as u64,
    );
    epoll_ctl_safe(epoll_fd, EpollOp::EpollCtlAdd, pid_fd, Some(&mut event))
        .context("Failed to add PIDFd to epoll")?;

    // Add listening socket to epoll (not necessary to set EPOLLERR).
    let mut event = EpollEvent::new(
        EpollFlags::EPOLLET | EpollFlags::EPOLLIN | EpollFlags::EPOLLHUP | EpollFlags::EPOLLRDHUP,
        listening_fd as u64,
    );
    epoll_ctl_safe(
        epoll_fd,
        EpollOp::EpollCtlAdd,
        listening_fd,
        Some(&mut event),
    )
    .context("Failed to add listening socket to epoll")?;

    let mut client_manager = ClientManager::new();

    // TODO: MAX_EVENTS=1024 move to config.rs
    let mut events = vec![EpollEvent::empty(); 1024];
    loop {
        // Wait for events and handle EINTR.
        let n = match epoll_wait(epoll_fd, &mut events, -1) {
            Ok(n) => n,
            Err(Errno::EINTR) => continue, // Retry if interrupted by a signal
            Err(e) => return Err(anyhow::Error::msg(e)).context("Failed to wait on epoll"),
        };

        for event in events.iter().take(n) {
            let fd = event.data() as RawFd;
            let event_flags = event.events();

            if fd == pid_fd {
                // Syd exited, exit gracefully.
                return Ok(());
            } else if fd == listening_fd {
                if event_flags.intersects(
                    EpollFlags::EPOLLHUP | EpollFlags::EPOLLRDHUP | EpollFlags::EPOLLERR,
                ) {
                    // The listening socket's other end has been closed
                    // or an error occurred, exit gracefully.
                    return Ok(());
                }

                // Accept new connection
                handle_new_connection(
                    listening_fd,
                    &*external_addr,
                    epoll_fd,
                    &mut client_manager,
                )?;
            } else {
                // Handle events for existing connections
                handle_existing_connection(fd, event_flags, epoll_fd, &mut client_manager)?;
            }
        }
    }
}

fn handle_existing_connection(
    fd: RawFd,
    event_flags: EpollFlags,
    epoll_fd: RawFd,
    client_manager: &mut ClientManager,
) -> Result<()> {
    if let Some(client) = client_manager.get_client_mut(fd) {
        let result = if event_flags.contains(EpollFlags::EPOLLIN) {
            // Handle readable events
            if fd == client.client_fd {
                client.handle_splice(false)
            } else if fd == client.ext_fd {
                client.handle_splice(true)
            } else {
                Ok(())
            }
        } else if event_flags.contains(EpollFlags::EPOLLOUT) {
            // Handle writable events
            if fd == client.client_fd {
                client.attempt_write(false)
            } else if fd == client.ext_fd {
                client.attempt_write(true)
            } else {
                Ok(())
            }
        } else if event_flags
            .intersects(EpollFlags::EPOLLERR | EpollFlags::EPOLLHUP | EpollFlags::EPOLLRDHUP)
        {
            // Handle disconnection events.
            Err(Errno::EPIPE)
        } else {
            Ok(())
        };

        // Handle errors and close clients if needed.
        if let Err(e) = result {
            if e != Errno::EAGAIN {
                if let Some(client) = client_manager.remove_client(fd) {
                    client.close(epoll_fd);
                }
            }
        }
    }

    Ok(())
}

fn handle_new_connection(
    listening_fd: RawFd,
    external_addr: &dyn SockaddrLike,
    epoll_fd: RawFd,
    client_manager: &mut ClientManager,
) -> Result<()> {
    #[allow(unreachable_patterns)]
    let client_fd = match accept4(
        listening_fd,
        SockFlag::SOCK_NONBLOCK | SockFlag::SOCK_CLOEXEC,
    ) {
        Ok(fd) => fd,
        Err(Errno::EAGAIN | Errno::EWOULDBLOCK) => return Ok(()), // No more connections to accept
        Err(e) => return Err(anyhow::Error::msg(e)).context("Failed to accept connection")?,
    };
    let client = Client::new(client_fd, external_addr)?;
    let ext_fd = client.ext_fd;
    client_manager.add_client(client);

    let mut event = EpollEvent::new(
        EpollFlags::EPOLLET
            | EpollFlags::EPOLLIN
            | EpollFlags::EPOLLOUT
            | EpollFlags::EPOLLHUP
            | EpollFlags::EPOLLRDHUP,
        client_fd as u64,
    );
    epoll_ctl_safe(epoll_fd, EpollOp::EpollCtlAdd, client_fd, Some(&mut event))
        .context("Failed to add client socket to epoll")?;

    let mut event = EpollEvent::new(
        EpollFlags::EPOLLET
            | EpollFlags::EPOLLIN
            | EpollFlags::EPOLLOUT
            | EpollFlags::EPOLLHUP
            | EpollFlags::EPOLLRDHUP,
        ext_fd as u64,
    );
    epoll_ctl_safe(epoll_fd, EpollOp::EpollCtlAdd, ext_fd, Some(&mut event))
        .context("Failed to add external socket to epoll")?;

    Ok(())
}

struct ClientManager {
    clients: HashMap<RawFd, Client>,
    fd_index: HashMap<RawFd, RawFd>,
}

impl ClientManager {
    fn new() -> Self {
        ClientManager {
            clients: HashMap::new(),
            fd_index: HashMap::new(),
        }
    }

    fn add_client(&mut self, client: Client) {
        let client_fd = client.client_fd;
        let ext_fd = client.ext_fd;
        self.clients.insert(client_fd, client);
        self.fd_index.insert(client_fd, client_fd);
        self.fd_index.insert(ext_fd, client_fd);
    }

    fn remove_client(&mut self, fd: RawFd) -> Option<Client> {
        if let Some(client_fd) = self.fd_index.remove(&fd) {
            self.fd_index.remove(&self.clients[&client_fd].ext_fd);
            return self.clients.remove(&client_fd);
        }
        None
    }

    fn get_client_mut(&mut self, fd: RawFd) -> Option<&mut Client> {
        if let Some(&client_fd) = self.fd_index.get(&fd) {
            self.clients.get_mut(&client_fd)
        } else {
            None
        }
    }
}

struct Client {
    client_fd: RawFd,
    pipe_in_from_client: RawFd,
    pipe_out_to_client: RawFd,
    pipe_in_from_ext: RawFd,
    pipe_out_to_ext: RawFd,
    ext_fd: RawFd,
}

impl Client {
    /// Create a new client
    fn new(client_fd: RawFd, external_addr: &dyn SockaddrLike) -> Result<Self> {
        let (pipe_in_from_client, pipe_out_to_client) = pipe2(OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)
            .map_err(anyhow::Error::msg)
            .context("Failed to create client pipe")?;
        let (pipe_in_from_ext, pipe_out_to_ext) = pipe2(OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)
            .map_err(anyhow::Error::msg)
            .context("Failed to create external pipe")?;

        let family = external_addr
            .family()
            .ok_or(Errno::EINVAL)
            .map_err(anyhow::Error::msg)
            .context("Invalid address family")?;
        let ext_fd = socket(
            family,
            SockType::Stream,
            SockFlag::SOCK_NONBLOCK | SockFlag::SOCK_CLOEXEC,
            None,
        )
        .map_err(anyhow::Error::msg)
        .context("Failed to create socket")?;

        connect(ext_fd, external_addr)
            .or_else(|e| {
                if e == Errno::EINPROGRESS {
                    Ok(())
                } else {
                    Err(e)
                }
            })
            .map_err(anyhow::Error::msg)
            .context("Failed to connect to external address")?;

        Ok(Client {
            client_fd,
            pipe_in_from_client,
            pipe_out_to_client,
            pipe_in_from_ext,
            pipe_out_to_ext,
            ext_fd,
        })
    }

    /// Close client connection and clean up resources
    fn close(self, epoll_fd: RawFd) {
        let _ = epoll_ctl_safe(epoll_fd, EpollOp::EpollCtlDel, self.client_fd, None);
        let _ = epoll_ctl_safe(epoll_fd, EpollOp::EpollCtlDel, self.ext_fd, None);
        let _ = close(self.client_fd);
        let _ = close(self.ext_fd);
        let _ = close(self.pipe_in_from_client);
        let _ = close(self.pipe_out_to_client);
        let _ = close(self.pipe_in_from_ext);
        let _ = close(self.pipe_out_to_ext);
    }

    fn handle_splice(&self, ext: bool) -> std::result::Result<(), Errno> {
        let (src_fd, dst_fd, pipe_in_fd, pipe_out_fd) = if ext {
            (
                self.ext_fd,
                self.client_fd,
                self.pipe_in_from_client,
                self.pipe_out_to_client,
            )
        } else {
            (
                self.client_fd,
                self.ext_fd,
                self.pipe_in_from_ext,
                self.pipe_out_to_ext,
            )
        };

        loop {
            match Self::splice_data(src_fd, pipe_out_fd) {
                Ok(_) => while Self::splice_data(pipe_in_fd, dst_fd).is_ok() {},
                Err(error) => break Err(error),
            }
        }
    }

    fn attempt_write(&self, ext: bool) -> std::result::Result<(), Errno> {
        let (src_fd, dst_fd) = if ext {
            (self.pipe_in_from_client, self.ext_fd)
        } else {
            (self.pipe_in_from_ext, self.client_fd)
        };

        loop {
            match Self::splice_data(src_fd, dst_fd) {
                Ok(_) => {}
                Err(error) => break Err(error),
            }
        }
    }

    /// Splice data from source to destination
    fn splice_data(src_fd: RawFd, dst_fd: RawFd) -> std::result::Result<(), Errno> {
        match splice(
            src_fd,
            None,
            dst_fd,
            None,
            PIPE_BUF,
            SpliceFFlags::SPLICE_F_NONBLOCK | SpliceFFlags::SPLICE_F_MORE,
        )? {
            0 => Err(Errno::EPIPE), // Handle EOF
            _ => Ok(()),
        }
    }
}

/// Safe wrapper for epoll_ctl with detailed error handling
fn epoll_ctl_safe(
    epoll_fd: RawFd,
    op: EpollOp,
    fd: RawFd,
    event: Option<&mut EpollEvent>,
) -> Result<()> {
    match epoll_ctl(epoll_fd, op, fd, event) {
        Ok(()) => Ok(()),
        Err(Errno::EEXIST) if op == EpollOp::EpollCtlAdd => Ok(()), // Ignore EEXIST for EPOLL_CTL_ADD
        Err(Errno::ENOENT) if op == EpollOp::EpollCtlDel => Ok(()), // Ignore ENOENT for EPOLL_CTL_DEL
        Err(e) => Err(anyhow::Error::msg(e)),
    }
}

/// Transit this process to a confined state.
fn confine(addr: &dyn SockaddrLike, dry_run: bool) -> Result<()> {
    let family = addr
        .family()
        .ok_or(Errno::EINVAL)
        .map_err(anyhow::Error::msg)
        .context("Invalid address family")?;
    let domain = match family {
        AddressFamily::Inet => nix::libc::AF_INET,
        AddressFamily::Inet6 => nix::libc::AF_INET6,
        _ => bail!("Invalid address family"),
    };
    let addr_ptr = addr.as_ptr() as u64;
    let addr_len = addr.len() as usize;

    // We add two seccomp filters:
    // 1. General filter that allows required syscalls.
    // 2. Mprotect filter that protects the whole memory region of the
    //    address pointer.
    let mut filter1 = new_filter(ScmpAction::KillProcess)?;
    let mut filter2 = new_filter(ScmpAction::Allow)?;

    let allow_call = [
        // can exit.
        "exit",
        "exit_group",
        // can allocate memory.
        // mmap{,2} and mprotect are further confined.
        "brk",
        "madvise",
        "munmap",
        "getrandom",
        // can handle signals.
        "sigaction",
        "sigaltstack",
        "sigpending",
        "sigprocmask",
        "sigsuspend",
        "sigreturn",
        "rt_sigaction",
        "rt_sigpending",
        "rt_sigprocmask",
        "rt_sigqueueinfo",
        "rt_sigreturn",
        "rt_sigtimedwait",
        "rt_sigtimedwait_time64",
        // can set file flags.
        "fcntl",
        // can close files.
        "close",
        // can do I/O with pipes.
        "pipe2",
        "splice",
        // can forward network.
        // socket and connect are further confined.
        "socketcall", // 32-bit compat.
        "accept4",
        // can use EPoll API,
        // can not create new EPoll FDs.
        "epoll_ctl",
        "epoll_wait",
        "epoll_pwait",
        "epoll_pwait2",
    ];

    // Default allowlist.
    for name in allow_call {
        filter1
            .add_rule(ScmpAction::Allow, ScmpSyscall::new(name))
            .map_err(anyhow::Error::msg)
            .with_context(|| format!("Failed to add seccomp allow rule for syscall {name}"))?;
    }

    // Restrict socket to the given domain, type and protocol.
    let sock_domain = domain as u64;
    let sock_type =
        (nix::libc::SOCK_STREAM | nix::libc::SOCK_NONBLOCK | nix::libc::SOCK_CLOEXEC) as u64;
    let sock_protocol = nix::libc::IPPROTO_IP as u64;
    filter1
        .add_rule_conditional_exact(
            ScmpAction::Allow,
            ScmpSyscall::new("socket"),
            &[
                scmp_cmp!($arg0 == sock_domain),
                scmp_cmp!($arg1 == sock_type),
                scmp_cmp!($arg2 == sock_protocol),
            ],
        )
        .map_err(anyhow::Error::msg)
        .context("Failed to add seccomp filter rule for syscall socket")?;

    // Restrict connect to a single safe pointer.
    filter1
        .add_rule_conditional_exact(
            ScmpAction::Allow,
            ScmpSyscall::new("connect"),
            &[scmp_cmp!($arg1 == addr_ptr)],
        )
        .map_err(anyhow::Error::msg)
        .context("Failed to add seccomp filter rule for syscall connect")?;

    // Prevent executable memory.
    const PROT_EXEC: u64 = nix::libc::PROT_EXEC as u64;
    for name in ["mmap", "mmap2", "mprotect"] {
        filter1
            .add_rule_conditional(
                ScmpAction::Allow,
                ScmpSyscall::new(name),
                &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
            )
            .map_err(anyhow::Error::msg)
            .with_context(|| format!("Failed to add seccomp filter rule for syscall {name}"))?;
    }

    // Allow write(2) to -42 for debugging.
    const MEANING_OF_LIFE: u64 = -42i64 as u64;
    filter1
        .add_rule_conditional(
            ScmpAction::Allow,
            ScmpSyscall::new("write"),
            &[scmp_cmp!($arg0 == MEANING_OF_LIFE)],
        )
        .map_err(anyhow::Error::msg)
        .context("Failed to add seccomp filter rule for syscall write")?;

    // Protect the memory area of address pointer,
    // from unwanted modifications.
    for offset in 0..=addr_len {
        let addr_ptr = addr_ptr.saturating_add(offset as u64);
        for name in ["mprotect", "munmap"] {
            filter2
                .add_rule_conditional(
                    ScmpAction::KillProcess,
                    ScmpSyscall::new(name),
                    &[scmp_cmp!($arg0 == addr_ptr)],
                )
                .map_err(anyhow::Error::msg)
                .with_context(|| {
                    "Failed to add seccomp filter for syscall mprotect at offset {offset}"
                })?;
        }
    }

    // Protect unwanted socketcall calls on 32-bit.
    // Ignore errors as socketcall may not be present.
    for call in 0..20 {
        if matches!(call, 0x1 | 0x3 | 0x12) {
            // socket | connect | accept4
            continue;
        }
        let _ = filter2.add_rule_conditional_exact(
            ScmpAction::KillProcess,
            ScmpSyscall::new("socketcall"),
            &[scmp_cmp!($arg0 == call)],
        );
    }

    if !dry_run {
        // Set up a Landlock sandbox to disallow all access.
        // Ignore errors as Landlock may not be supported.
        let p = vec![];
        let _ = syd::landlock_operation(&p, &p, syd::config::LANDLOCK_ABI);

        // Set up Memory-Deny-Write-Execute protections.
        // Ignore errors as PR_SET_MDWE may not be supported.
        const PR_SET_MDWE: nix::libc::c_int = 65;
        const PR_MDWE_REFUSE_EXEC_GAIN: nix::libc::c_ulong = 1;
        // SAFETY: nix has no wrapper for PR_SET_MDWE.
        let _ = unsafe { nix::libc::prctl(PR_SET_MDWE, PR_MDWE_REFUSE_EXEC_GAIN, 0, 0, 0) };

        // Set the process dumpable attribute to not-dumpable.
        // SAFETY: Our nix version does not have a wrapper for set_dumpable yet.
        let _ = unsafe { nix::libc::prctl(nix::libc::PR_SET_DUMPABLE, 0, 0, 0, 0) };
    }

    if dry_run {
        // Dump filter and return without loading.
        let _ = filter2.export_pfc(&mut std::io::stderr());
        let _ = filter1.export_pfc(&mut std::io::stderr());
    } else {
        // All done, load seccomp filter and begin confinement.
        // Careful, we have to load the mprotect filter first,
        // as the first general filter is more restrictive.
        filter2
            .load()
            .map_err(anyhow::Error::msg)
            .context("Failed to load the mprotect seccomp filter")?;
        filter1
            .load()
            .map_err(anyhow::Error::msg)
            .context("Failed to load the general seccomp filter")?;
    }

    Ok(())
}

fn new_filter(action: ScmpAction) -> Result<ScmpFilterContext> {
    let mut filter = ScmpFilterContext::new_filter(action)
        .map_err(anyhow::Error::msg)
        .context("Failed to create seccop filter")?;

    // Enforce the NO_NEW_PRIVS functionality before
    // loading the seccomp filter into the kernel.
    filter
        .set_ctl_nnp(true)
        .map_err(anyhow::Error::msg)
        .context("Failed to enforce no new privs")?;

    // Kill process for bad arch.
    filter
        .set_act_badarch(ScmpAction::KillProcess)
        .map_err(anyhow::Error::msg)
        .context("Failed to set seccomp bad architecture action")?;

    // Use a binary tree sorted by syscall number, if possible.
    let _ = filter.set_ctl_optimize(2);

    Ok(filter)
}

// Parse command line options.
#[allow(clippy::type_complexity)]
fn parse_options() -> Result<(OwnedFd, UnixStream, Box<dyn SockaddrLike>, bool)> {
    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
    let mut port = 9050;
    let mut sock: Option<&str> = None;
    let mut fpid: Option<&str> = None;
    let mut debug = false; // If enabled, skips confinement.

    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                exit(0);
            }
            Opt::Short('d') => {
                debug = true;
            }
            Opt::Short('p') => {
                fpid = opts.value().ok();
            }
            Opt::Short('i') => {
                sock = opts.value().ok();
            }
            Opt::Short('o') => {
                (addr, port) = match opts.value() {
                    Ok(value) => {
                        let (mut addr, port) = value
                            .rsplit_once(':')
                            .ok_or(anyhow::anyhow!("Address missing ':'"))?;
                        // if the address contains a ':', it should be parsed as an ipv6 address
                        if addr.contains(':') {
                            // strip the surrounding square brackets (for example '[::1]')
                            addr = addr
                                .strip_prefix('[')
                                .ok_or(anyhow::anyhow!(
                                    "IPv6 address missing surrounding bracket '['"
                                ))?
                                .strip_suffix(']')
                                .ok_or(anyhow::anyhow!(
                                    "IPv6 address missing surrounding bracket ']'"
                                ))?;
                        }

                        // parse the address
                        let addr: IpAddr = addr
                            .parse()
                            .map_err(anyhow::Error::msg)
                            .with_context(|| format!("Failed to parse address '{addr}'"))?;

                        // parse the external port
                        let port: u16 = port
                            .parse()
                            .map_err(anyhow::Error::msg)
                            .with_context(|| format!("Failed to parse port '{port}'"))?;

                        (addr, port)
                    }
                    Err(_) => {
                        eprintln!("Error: -o requires an argument.");
                        help();
                        exit(1);
                    }
                };
            }
            _ => {
                eprintln!("Unknown option: {opt:?}");
                help();
                exit(1);
            }
        }
    }

    let fpid = if let Some(fpid) = fpid {
        // Parse file descriptor.
        let fpid = fpid
            .parse::<RawFd>()
            .map_err(anyhow::Error::msg)
            .with_context(|| format!("Failed to parse PIDFd '{fpid}'"))?;

        if fpid < 0 {
            bail!("Invalid PIDFd '{fpid}'");
        }

        // Validate file descriptor.
        let mut buf = itoa::Buffer::new();
        let mut pfd = PathBuf::from("/proc/self/fd");
        pfd.push(buf.format(fpid));
        if !path_exists(&pfd) {
            bail!("No such PIDFd '{fpid}'");
        }

        // SAFETY: Syd passes a valid PID FD to syd-tor.
        // Any other usecase is unsupported.
        unsafe { OwnedFd::from_raw_fd(fpid) }
    } else {
        eprintln!("Error: -p is required.");
        help();
        exit(1);
    };

    let sock = if let Some(sock) = sock {
        // Parse file descriptor.
        let sock = sock
            .parse::<RawFd>()
            .map_err(anyhow::Error::msg)
            .with_context(|| format!("Failed to parse socket '{sock}'"))?;

        if sock < 0 {
            bail!("Invalid socket '{sock}'");
        }

        // Validate file descriptor.
        let mut buf = itoa::Buffer::new();
        let mut pfd = PathBuf::from("/proc/self/fd");
        pfd.push(buf.format(sock));
        if !path_exists(&pfd) {
            bail!("No such socket '{sock}'");
        }

        // SAFETY: Syd passes a valid socket FD to syd-tor.
        // Any other usecase is unsupported.
        unsafe { UnixStream::from_raw_fd(sock) }
    } else {
        eprintln!("Error: -i is required.");
        help();
        exit(1);
    };

    // Validate socket address and convert.
    let addr = SocketAddr::new(addr, port);
    let addr: Box<dyn SockaddrLike> = match addr {
        SocketAddr::V4(addr) => Box::new(SockaddrIn::from(addr)),
        SocketAddr::V6(addr) => Box::new(SockaddrIn6::from(addr)),
    };

    Ok((fpid, sock, addr, debug))
}

fn help() {
    println!("Usage: syd-tor [-h] -p <pid-fd> -i <socket-fd> [-o addr:port]");
    println!("Syd's SOCKS Proxy Forwarder");
    println!("Receives listening socket from fd and forwards traffic to addr:port.");
    println!("External address must be an IPv4 or IPv6 address.");
    println!("PID file descriptor is used to track the exit of Syd process.");
    println!("  -h             Print this help message and exit.");
    println!("  -d             Run in debug mode without confinement.");
    println!("  -p <pid-fd>    PID file descriptor of Syd process.");
    println!("  -i <socket-fd> Socket file descriptor to receive the listening socket from.");
    println!("  -o <addr:port> Specify external address to forward traffic to.");
    println!("                 Defaults to 127.0.0.1:9050.");
}
